﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

namespace LarrysArms
{
	public class HairColorChanger : MVRScript
	{
		Atom _person;
		List<DAZHairGroup> _activeHairGroups;

		UIDynamicButton _btnUpdateHairList;
		UIDynamicButton _btnApplyToAll;
		UIDynamicButton _btnRestoreHairDefaults;
		UIDynamicButton _btnPresetBrowse;

		JSONStorableUrl _presetFileUrl;
		JSONStorableStringChooser _chooserHairStorable;

		JSONStorableFloat _sliderPrimarySpecGloss;
		JSONStorableFloat _sliderSecondarySpecGloss;
		JSONStorableFloat _sliderSecondarySpecShift;
		JSONStorableFloat _sliderRandomColorPower;
		JSONStorableFloat _sliderRandomColorOffset;
		JSONStorableFloat _sliderDiffuseSoftness;
		JSONStorableFloat _sliderFresnelPower;
		JSONStorableFloat _sliderFresnelAttenuation;
		JSONStorableFloat _sliderGiFactor;
		JSONStorableFloat _sliderNormalRandomize;

		JSONStorableColor _colorPickerRoot;
		JSONStorableColor _colorPickerTip;
		JSONStorableFloat _sliderRootTipRolloff;
		JSONStorableColor _colorPickerSpec;

		UIDynamicButton _btnMaterialSliderDefaults;
		UIDynamicButton _btnMaterialColorDefaults;


		public override void Init()
		{
			try
			{
				_person = SuperController.singleton.GetSelectedAtom();
				if (_person.category != "People")
				{
					SuperController.LogError("The plugin must be on a Person atom.");
					return;
				}

				UpdateActiveHairGroups();   // Update hair list
				CreateUI(); // Create UI
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}
		}

		protected void UpdateActiveHairGroups()
		{
			_activeHairGroups = new List<DAZHairGroup>();

			foreach (DAZHairGroup hairGroup in _person.GetComponentsInChildren<DAZHairGroup>())
			{
				DAZHairGroupControl hairControl = hairGroup.GetComponentInChildren<DAZHairGroupControl>();
				HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

				if (hairControl != null && simControl != null)
				{
					_activeHairGroups.Add(hairGroup);
				}
			}
		}

		protected void CreateUI()
		{
			// Update Hair List button
			_btnUpdateHairList = CreateButton("Update Hair List");
			_btnUpdateHairList.button.onClick.AddListener(UpdateHairListCallback);

			// Hair chooser drop-down
			_chooserHairStorable = new JSONStorableStringChooser("hairStorableChooser", GetHairNames(), "None", "Values from: ", HairStorableChooserCallback);
			UIDynamicPopup udp = CreatePopup(_chooserHairStorable);

			// Apply To All button
			_btnApplyToAll = CreateButton("Apply To All");
			_btnApplyToAll.button.onClick.AddListener(ApplyToAllCallback);

			// Restore Hair Defaults button
			_btnRestoreHairDefaults = CreateButton("Restore Hair Defaults");
			_btnRestoreHairDefaults.button.onClick.AddListener(RestoreHairDefaultsCallback);

			// Plugin info text field
			JSONStorableString stringVersion = new JSONStorableString("versionNum", "Hair Color Changer 1.1 by LarrysArms.");
			UIDynamicTextField textFieldVersion = CreateTextField(stringVersion, true);
			textFieldVersion.height = 210;

			// Preset file browser
			var presetPath = Application.dataPath.Replace("/VaM_Data", "") + "/Custom/Hair/Female/";
			_btnPresetBrowse = CreateButton("Browse Presets ", true);
			_presetFileUrl = new JSONStorableUrl("Preset Browser", presetPath, FileBrowserCallback, "vap");
			_presetFileUrl.suggestedPath = presetPath;

			_presetFileUrl.RegisterFileBrowseButton(_btnPresetBrowse.button);

			CreateSpacer().height = 50;
			CreateSpacer(true).height = 50;

			// Material info properties
			_sliderPrimarySpecGloss = new JSONStorableFloat("Primary Specular Gloss", 160.00f, SliderCallback, 0, 256.00f);
			_sliderSecondarySpecGloss = new JSONStorableFloat("Secondary Specular Gloss", 64.00f, SliderCallback, 0, 256.0f);
			_sliderSecondarySpecShift = new JSONStorableFloat("Secondary Specular Shift", 0.400f, SliderCallback, 0, 1.0f);
			_sliderRandomColorPower = new JSONStorableFloat("Random Color Power", 2.000f, SliderCallback, 0, 10.0f);
			_sliderRandomColorOffset = new JSONStorableFloat("Random Color Offset", 0.300f, SliderCallback, 0, 1.0f);
			_sliderDiffuseSoftness = new JSONStorableFloat("Diffuse Softness", 0.100f, SliderCallback, 0, 1.0f);
			_sliderFresnelPower = new JSONStorableFloat("Fresnel Power", 8.00f, SliderCallback, 0, 10.0f);
			_sliderFresnelAttenuation = new JSONStorableFloat("Fresnel Attenuation", 0.200f, SliderCallback, 0, 1.0f);
			_sliderGiFactor = new JSONStorableFloat("GI Factor", 0.500f, SliderCallback, 0, 1.0f);
			_sliderNormalRandomize = new JSONStorableFloat("Normal Randomize", 0.00f, SliderCallback, 0, 1.0f);

			_sliderRootTipRolloff = new JSONStorableFloat("Root Tip Rolloff", 3f, SliderCallback, 0f, 5f, true);

			HSVColor rootColor; rootColor.H = 0.079f; rootColor.S = 0.91f; rootColor.V = 0.09f;
			_colorPickerRoot = new JSONStorableColor("Root Color", rootColor, ColorPickerCallback);
			HSVColor tipColor; tipColor.H = 0.079f; tipColor.S = 0.68f; tipColor.V = 0.57f;
			_colorPickerTip = new JSONStorableColor("Tip Color", tipColor, ColorPickerCallback);
			HSVColor specColor; specColor.H = 0f; specColor.S = 0f; specColor.V = 0.67f;
			_colorPickerSpec = new JSONStorableColor("Spec Color", specColor, ColorPickerCallback);

			RegisterColor(_colorPickerRoot);
			RegisterColor(_colorPickerTip);
			RegisterColor(_colorPickerSpec);

			// Material info left-hand side
			CreateSlider(_sliderPrimarySpecGloss);
			CreateSlider(_sliderSecondarySpecGloss);
			CreateSlider(_sliderSecondarySpecShift);
			CreateSlider(_sliderRandomColorPower);
			CreateSlider(_sliderRandomColorOffset);
			CreateSlider(_sliderDiffuseSoftness);
			CreateSlider(_sliderFresnelPower);
			CreateSlider(_sliderFresnelAttenuation);
			CreateSlider(_sliderGiFactor);
			CreateSlider(_sliderNormalRandomize);

			// Material info right-hand side
			CreateColorPicker(_colorPickerRoot, true);
			CreateColorPicker(_colorPickerTip, true);
			CreateSlider(_sliderRootTipRolloff, true);
			CreateColorPicker(_colorPickerSpec, true);

			CreateSpacer(true).height = 15;

			// Reset Material Slider defaults button
			_btnMaterialSliderDefaults = CreateButton("Reset Material Slider defaults");
			_btnMaterialSliderDefaults.button.onClick.AddListener(ResetMaterialSliderCallback);

			// Reset Color Picker defaults button
			_btnMaterialColorDefaults = CreateButton("Reset Color Picker defaults", true);
			_btnMaterialColorDefaults.button.onClick.AddListener(ResetColorPickerCallback);
		}

		protected void FileBrowserCallback(string file)
		{
			var json = SuperController.singleton.LoadJSON(file);

			JSONNode hairProperties = json["storables"][1];
			JSONNode rootColorNode = hairProperties["rootColor"];
			JSONNode tipColorNode = hairProperties["tipColor"];
			JSONNode specColorNode = hairProperties["specularColor"];

			// TODO - Improve this!
			HSVColor rootColor; rootColor.H = rootColorNode["h"].AsFloat; rootColor.S = rootColorNode["s"].AsFloat; rootColor.V = rootColorNode["v"].AsFloat;
			HSVColor tipColor; tipColor.H = tipColorNode["h"].AsFloat; tipColor.S = tipColorNode["s"].AsFloat; tipColor.V = tipColorNode["v"].AsFloat;
			HSVColor specColor; specColor.H = specColorNode["h"].AsFloat; specColor.S = specColorNode["s"].AsFloat; specColor.V = specColorNode["v"].AsFloat;

			_sliderPrimarySpecGloss.val = hairProperties["primarySpecularSharpness"].AsFloat;
			_sliderSecondarySpecGloss.val = hairProperties["secondarySpecularSharpness"].AsFloat;
			_sliderSecondarySpecShift.val = hairProperties["specularShift"].AsFloat;
			_sliderRandomColorPower.val = hairProperties["randomColorPower"].AsFloat;
			_sliderRandomColorOffset.val = hairProperties["randomColorOffset"].AsFloat;
			_sliderDiffuseSoftness.val = hairProperties["diffuseSoftness"].AsFloat;
			_sliderFresnelPower.val = hairProperties["fresnelPower"].AsFloat;
			_sliderFresnelAttenuation.val = hairProperties["fresnelAttenuation"].AsFloat;
			_sliderGiFactor.val = hairProperties["IBLFactor"].AsFloat;
			_sliderNormalRandomize.val = hairProperties["normalRandomize"].AsFloat;

			_colorPickerRoot.val = rootColor;
			_colorPickerTip.val = tipColor;
			_sliderRootTipRolloff.val = hairProperties["colorRolloff"].AsFloat;
			_colorPickerSpec.val = specColor;

			UpdateScalpMaterials(json);
		}

		protected void UpdateScalpMaterials(JSONNode jsonInfo)
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				DAZSkinWrapMaterialOptions scalpControl = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>();

				float currentScalpAlpha = scalpControl.GetFloatParamValue("Alpha Adjust");

				// Don't update anything if the current alpha is almost transparent.
				if (currentScalpAlpha > -0.97f)
				{
					var scalpProperties = jsonInfo["storables"][5];

					float specularTextureOffset = scalpProperties["Specular Texture Offset"].AsFloat;
					float specularIntensity = scalpProperties["Specular Intensity"].AsFloat;
					float gloss = scalpProperties["Gloss"].AsFloat;
					float specularFresnel = scalpProperties["Specular Fresnel"].AsFloat;
					float glossTextureOffset = scalpProperties["Gloss Texture Offset"].AsFloat;
					float giFilter = scalpProperties["Global Illumination Filter"].AsFloat;
					float alphaAdjust = scalpProperties["Alpha Adjust"].AsFloat;
					float diffuseTextureOffset = scalpProperties["Diffuse Texture Offset"].AsFloat;

					JSONNode diffuseColorNode = scalpProperties["Diffuse Color"];
					JSONNode specularColorNode = scalpProperties["Specular Color"];
					JSONNode subsurfaceColorNode = scalpProperties["Subsurface Color"];

					HSVColor diffuseColor; diffuseColor.H = diffuseColorNode["h"].AsFloat; diffuseColor.S = diffuseColorNode["s"].AsFloat; diffuseColor.V = diffuseColorNode["v"].AsFloat;
					HSVColor specularColor; specularColor.H = specularColorNode["h"].AsFloat; specularColor.S = specularColorNode["s"].AsFloat; specularColor.V = specularColorNode["v"].AsFloat;
					HSVColor subsurfaceColor; subsurfaceColor.H = subsurfaceColorNode["h"].AsFloat; subsurfaceColor.S = subsurfaceColorNode["s"].AsFloat; subsurfaceColor.V = subsurfaceColorNode["v"].AsFloat;

					scalpControl.SetFloatParamValue("Specular Texture Offset", specularTextureOffset);
					scalpControl.SetFloatParamValue("Specular Intensity", specularIntensity);
					scalpControl.SetFloatParamValue("Gloss", gloss);
					scalpControl.SetFloatParamValue("Specular Fresnel", specularFresnel);
					scalpControl.SetFloatParamValue("Gloss Texture Offset", glossTextureOffset);
					scalpControl.SetFloatParamValue("Global Illumination Filter", giFilter);
					scalpControl.SetFloatParamValue("Alpha Adjust", alphaAdjust);
					scalpControl.SetFloatParamValue("Diffuse Texture Offset", diffuseTextureOffset);

					scalpControl.SetColorParamValue("Diffuse Color", diffuseColor);
					scalpControl.SetColorParamValue("Specular Color", specularColor);
					scalpControl.SetColorParamValue("Subsurface Color", subsurfaceColor);
				}
			}
		}

		protected List<string> GetHairNames()
		{
			List<string> hairNames = new List<string>();

			foreach (DAZHairGroup hairGroup in _activeHairGroups)
			{
				hairNames.Add(hairGroup.displayName);
			}

			return hairNames;
		}

		public void UpdateHairListCallback()
		{
			UpdateActiveHairGroups();
			_chooserHairStorable.SetVal("");
			_chooserHairStorable.choices = GetHairNames();
		}

		public void ApplyToAllCallback()
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

				if (simControl != null)
				{
					SetCurrentMaterialProperties(simControl);
				}
			}
		}

		public void RestoreHairDefaultsCallback()
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();
				DAZSkinWrapMaterialOptions scalpControl = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>();

				if (simControl != null && scalpControl != null)
				{
					_chooserHairStorable.SetVal("");
					simControl.RestoreAllFromDefaults();
					scalpControl.RestoreAllFromDefaults();
					GetCurrentMaterialProperties(simControl);
				}
			}
		}

		protected void SliderCallback(float value)
		{
			ApplyToAllCallback();
		}

		protected void ColorPickerCallback(JSONStorableColor jcolor)
		{
			ApplyToAllCallback();
		}

		protected void ResetMaterialSliderCallback()
		{
			_sliderPrimarySpecGloss.defaultVal = 160f;
			_sliderSecondarySpecGloss.defaultVal = 64f;
			_sliderSecondarySpecShift.defaultVal = 0.4f;
			_sliderRandomColorPower.defaultVal = 2.0f;
			_sliderRandomColorOffset.defaultVal = 0.3f;
			_sliderDiffuseSoftness.defaultVal = 0.1f;
			_sliderFresnelPower.defaultVal = 8.0f;
			_sliderFresnelAttenuation.defaultVal = 0.2f;
			_sliderGiFactor.defaultVal = 0.5f;
			_sliderNormalRandomize.defaultVal = 0.0f;

			_sliderRootTipRolloff.defaultVal = 3.0f;

			_sliderPrimarySpecGloss.SetValToDefault();
			_sliderSecondarySpecGloss.SetValToDefault();
			_sliderSecondarySpecShift.SetValToDefault();
			_sliderRandomColorPower.SetValToDefault();
			_sliderRandomColorOffset.SetValToDefault();
			_sliderDiffuseSoftness.SetValToDefault();
			_sliderFresnelPower.SetValToDefault();
			_sliderFresnelAttenuation.SetValToDefault();
			_sliderGiFactor.SetValToDefault();
			_sliderNormalRandomize.SetValToDefault();

			_sliderRootTipRolloff.SetValToDefault();
		}

		protected void ResetColorPickerCallback()
		{
			HSVColor rootColor; rootColor.H = 0.079f; rootColor.S = 0.91f; rootColor.V = 0.09f;
			HSVColor tipColor; tipColor.H = 0.079f; tipColor.S = 0.68f; tipColor.V = 0.57f;
			HSVColor specColor; specColor.H = 0f; specColor.S = 0f; specColor.V = 0.67f;

			_colorPickerRoot.defaultVal = rootColor;
			_colorPickerTip.defaultVal = tipColor;
			_colorPickerSpec.defaultVal = specColor;

			_colorPickerRoot.SetValToDefault();
			_colorPickerTip.SetValToDefault();
			_colorPickerSpec.SetValToDefault();
		}

		protected void HairStorableChooserCallback(string hairStorableID)
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				if (hairStorableID == hairGroup.displayName)
				{
					HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

					if (simControl != null)
					{
						GetCurrentMaterialProperties(simControl);

						// Set slider defaults to material defaults
						_sliderPrimarySpecGloss.SetDefaultFromCurrent();
						_sliderSecondarySpecGloss.SetDefaultFromCurrent();
						_sliderSecondarySpecShift.SetDefaultFromCurrent();
						_sliderRandomColorPower.SetDefaultFromCurrent();
						_sliderRandomColorOffset.SetDefaultFromCurrent();
						_sliderDiffuseSoftness.SetDefaultFromCurrent();
						_sliderFresnelPower.SetDefaultFromCurrent();
						_sliderFresnelAttenuation.SetDefaultFromCurrent();
						_sliderGiFactor.SetDefaultFromCurrent();
						_sliderNormalRandomize.SetDefaultFromCurrent();

						_colorPickerRoot.SetDefaultFromCurrent();
						_colorPickerTip.SetDefaultFromCurrent();
						_sliderRootTipRolloff.SetDefaultFromCurrent();
						_colorPickerSpec.SetDefaultFromCurrent();
					}
				}
			}
		}

		protected void SetCurrentMaterialProperties(HairSimControl hairControl)
		{
			hairControl.SetFloatParamValue("primarySpecularSharpness", _sliderPrimarySpecGloss.val);
			hairControl.SetFloatParamValue("secondarySpecularSharpness", _sliderSecondarySpecGloss.val);
			hairControl.SetFloatParamValue("specularShift", _sliderSecondarySpecShift.val);
			hairControl.SetFloatParamValue("randomColorPower", _sliderRandomColorPower.val);
			hairControl.SetFloatParamValue("randomColorOffset", _sliderRandomColorOffset.val);
			hairControl.SetFloatParamValue("diffuseSoftness", _sliderDiffuseSoftness.val);
			hairControl.SetFloatParamValue("fresnelPower", _sliderFresnelPower.val);
			hairControl.SetFloatParamValue("fresnelAttenuation", _sliderFresnelAttenuation.val);
			hairControl.SetFloatParamValue("IBLFactor", _sliderGiFactor.val);
			hairControl.SetFloatParamValue("normalRandomize", _sliderNormalRandomize.val);

			hairControl.SetColorParamValue("rootColor", _colorPickerRoot.val);
			hairControl.SetColorParamValue("tipColor", _colorPickerTip.val);
			hairControl.SetFloatParamValue("colorRolloff", _sliderRootTipRolloff.val);
			hairControl.SetColorParamValue("specularColor", _colorPickerSpec.val);
		}

		protected void GetCurrentMaterialProperties(HairSimControl simControl)
		{
			_sliderPrimarySpecGloss.valNoCallback = simControl.GetFloatParamValue("primarySpecularSharpness");
			_sliderSecondarySpecGloss.valNoCallback = simControl.GetFloatParamValue("secondarySpecularSharpness");
			_sliderSecondarySpecShift.valNoCallback = simControl.GetFloatParamValue("specularShift");
			_sliderRandomColorPower.valNoCallback = simControl.GetFloatParamValue("randomColorPower");
			_sliderRandomColorOffset.valNoCallback = simControl.GetFloatParamValue("randomColorOffset");
			_sliderDiffuseSoftness.valNoCallback = simControl.GetFloatParamValue("diffuseSoftness");
			_sliderFresnelPower.valNoCallback = simControl.GetFloatParamValue("fresnelPower");
			_sliderFresnelAttenuation.valNoCallback = simControl.GetFloatParamValue("fresnelAttenuation");
			_sliderGiFactor.valNoCallback = simControl.GetFloatParamValue("IBLFactor");
			_sliderNormalRandomize.valNoCallback = simControl.GetFloatParamValue("normalRandomize");

			_colorPickerRoot.valNoCallback = simControl.GetColorParamValue("rootColor");
			_colorPickerTip.valNoCallback = simControl.GetColorParamValue("tipColor");
			_sliderRootTipRolloff.valNoCallback = simControl.GetFloatParamValue("colorRolloff");
			_colorPickerSpec.valNoCallback = simControl.GetColorParamValue("specularColor");
		}
	}
}